"""Main module for check_dependencies."""

from __future__ import annotations

import ast
import logging
from os.path import commonpath
from pathlib import Path
from typing import TYPE_CHECKING

from check_dependencies.builtin_module import BUILTINS
from check_dependencies.lib import AppConfig, Dependency, PyProjectToml, pkg

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable, Iterator, Sequence

logger = logging.getLogger("check_dependencies")
ERR_MISSING_DEPENDENCY = 2
ERR_EXTRA_DEPENDENCY = 4
ERR_NO_PYPROJECT = 8


def yield_wrong_imports(
    file_names: Sequence[str],
    app_cfg: AppConfig,
) -> Generator[str, None, int]:
    """Yield output lines of missing/unused imports.

    If cfg.show_all is True, all imports are shown.
    """
    used_deps: set[str] = set()
    src_fmt = app_cfg.mk_src_formatter()
    try:
        pyproject_candidate = Path(
            commonpath(Path(p).expanduser().resolve() for p in file_names),
        )
    except ValueError as exc:
        logger.fatal("Could not find pyproject.toml in common path: %s", exc)
        return ERR_NO_PYPROJECT

    src_cfg = PyProjectToml.from_pyproject(pyproject_candidate, app_cfg=app_cfg)
    exit_status = 0

    expected_dependencies = frozenset().union(
        src_cfg.dependencies,  # dependencies from pyproject.toml
        BUILTINS,  # builtins
    )
    allowed_dependencies = frozenset().union(
        expected_dependencies,
        app_cfg.known_missing,
        src_cfg.known_missing,
    )

    for src_pth in _collect_files(file_names):
        for cause, module, stmt in _missing_imports_iter(
            src_pth,
            dependencies=allowed_dependencies,
        ):
            if cause != Dependency.OK:
                exit_status |= ERR_MISSING_DEPENDENCY
            used_deps.add(pkg(module))
            yield from src_fmt(src_pth, cause, module, stmt)

    if superfluous_requirements := [
        msg
        for dep in sorted(
            src_cfg.dependencies.difference(
                used_deps,
                app_cfg.known_extra,
                src_cfg.known_extra,
            ),
        )
        for msg in app_cfg.unused_fmt(dep)
    ]:
        exit_status |= ERR_EXTRA_DEPENDENCY
        if app_cfg.verbose:
            yield ""
            yield "### Dependencies in config file not used in application:"
            yield f"# Config file: {src_cfg.file}"
        yield from superfluous_requirements
    return exit_status


def _collect_files(file_names: Sequence[str]) -> Iterator[Path]:
    """Collect all Python files in a list of files or directories.

    Ensure no file is visited more than once.
    """
    visited = set()
    for p in map(Path, file_names):
        for p_sub in p.rglob("*.py") if p.is_dir() else [p]:
            if (p_sub_resolved := p_sub.resolve()) in visited:
                continue
            visited.add(p_sub_resolved)
            yield p_sub


def _missing_imports_iter(
    file: Path,
    dependencies: Iterable[str],
) -> Iterator[tuple[Dependency, str, ast.stmt]]:
    """Find missing imports in a Python file.

    :param file: Pyton file to analyse
    :param dependencies: Declared dependencies from pyproject file
    :param seen: Cache of seen packages - this is changed during the iteration
    :yields: Tuple of status, module name and optional import statement
    """
    try:
        parsed = ast.parse(file.read_text(), filename=str(file))
    except SyntaxError:
        logger.exception("Could not parse %s", file)
        return
    for module, stmt in _imports_iter(parsed.body):
        pkg_ = pkg(module)
        status = Dependency.OK if pkg_ in dependencies else Dependency.NA
        yield status, module, stmt


def _imports_iter(body: list[ast.stmt]) -> Iterator[tuple[str, ast.stmt]]:
    """Yield all import statements from a body of code."""
    for x in body:
        if isinstance(x, ast.Import):
            for alias in x.names:
                yield alias.name, x  # yield x, not alias to get lineno
        elif isinstance(x, ast.ImportFrom) and x.level == 0:
            # level > 0 means relative import
            yield x.module or "", x
        elif hasattr(x, "body"):
            yield from _imports_iter(x.body)



"""Library for check_dependencies."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from enum import Enum
from itertools import chain, takewhile
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, TypeVar, cast

try:
    import tomllib
except ImportError:
    import toml as tomllib  # type: ignore[no-redef]

if TYPE_CHECKING:
    import ast
    from collections.abc import Iterable, Iterator


logger = logging.getLogger("check_dependencies.lib")
_PYPROJECT_TOML = Path("pyproject.toml")


class Dependency(Enum):
    """Possible dependency state."""

    NA = "!"  # Not Available
    EXTRA = "+"  # Extra dependency in config file
    OK = " "  # Correct import (declared in config file)


@dataclass()
class AppConfig:
    """Application config and helper functions."""

    include_dev: bool = False
    verbose: bool = False
    show_all: bool = False
    known_extra: Iterable[str] = frozenset()
    known_missing: Iterable[str] = frozenset()

    def __post_init__(self) -> None:
        """Dataclass post init method to ensure sets are frozen."""
        self.known_extra = frozenset(filter(None, self.known_extra or []))
        self.known_missing = frozenset(filter(None, self.known_missing or []))

    def mk_src_formatter(
        self,
    ) -> Callable[[Path | str, Dependency, str, ast.stmt | None], Iterator[str]]:
        """Formatter for missing or used dependencies."""
        cache: set[str] = set()

        def src_cause_formatter(
            src_pth: Path | str,
            cause: Dependency,
            module: str,
            stmt: ast.stmt | None,
        ) -> Iterator[str]:
            if self.verbose:
                location = f"{Path(src_pth).as_posix()}:{getattr(stmt, 'lineno', -1)}"
                if cause == Dependency.NA or self.show_all:
                    yield f"{cause.value}{cause.name} {location} {module}"
            elif (pkg_ := pkg(module)) not in cache:
                cache.add(pkg_)
                if cause == Dependency.NA or self.show_all:
                    yield f"{cause.value} {pkg_}"

        return src_cause_formatter

    def unused_fmt(self, module: str) -> Iterator[str]:
        """Formatter for unused but declared dependencies."""
        info = Dependency.EXTRA.name if self.verbose else ""
        yield f"{Dependency.EXTRA.value}{info} {module}"


@dataclass(frozen=True)
class PyProjectToml:
    """Project specific options (dependencies, config) from a pyproject.toml file."""

    file: Path
    cfg: dict[str, Any]
    include_dev: bool = False

    @classmethod
    def from_pyproject(cls, file: Path, app_cfg: AppConfig) -> PyProjectToml:
        """Create a PyProjectToml instance from a pyproject.toml file."""
        pyproject_file = _get_pyproject_path(file)
        return cls(
            file=pyproject_file,
            cfg=tomllib.loads(pyproject_file.read_text("utf-8")),
            include_dev=app_cfg.include_dev,
        )

    @property
    def dependencies(self) -> frozenset[str]:
        """Get dependencies from pyproject.toml file."""
        if "dependencies" in self.cfg.get("project", {}):
            return self._pep631_dependencies()
        if "poetry" in self.cfg.get("tool", {}):
            return self._poetry_dependencies()

        logger.warning("No dependencies found in %s", _PYPROJECT_TOML)
        return frozenset()

    @property
    def known_missing(self) -> frozenset[str]:
        """Known to be used in application but not declared in requirements."""
        # Add project name
        pep631_name = pkg(_nested_item(self.cfg, "project.name", str) or "")
        poetry_name = pkg(_nested_item(self.cfg, "tool.poetry.name", str) or "")
        return frozenset(
            filter(
                lambda x: x,
                [
                    pep631_name,
                    poetry_name,
                    *_nested_item(
                        self.cfg,
                        "tool.check-dependencies.known-missing",
                        list,
                    ),
                ],
            ),
        )

    @property
    def known_extra(self) -> frozenset[str]:
        """Dependencies that are known to be unused in application."""
        return frozenset(
            _nested_item(self.cfg, "tool.check-dependencies.known-extra", list),
        )

    def _poetry_dependencies(self) -> frozenset[str]:
        """Get dependencies from a poetry-style pyproject.toml file."""
        deps = set(_nested_item(self.cfg, "tool.poetry.dependencies", dict))
        if self.include_dev:
            deps |= set(
                _nested_item(self.cfg, "tool.poetry.group.dev.dependencies", dict),
            )
            deps |= set(_nested_item(self.cfg, "tool.poetry.dev-dependencies", dict))

        return frozenset(x for x in deps) - {"python"}

    def _pep631_dependencies(self) -> frozenset[str]:
        """Get dependencies from a PEP 631-style pyproject.toml file."""

        def canonical(name: str) -> str:
            return "".join(takewhile(lambda x: x.isalnum() or x in "-_", name)).replace(
                "-",
                "_",
            )

        raw_deps = _nested_item(self.cfg, "project.dependencies", list)
        deps = {canonical(raw_dep) for raw_dep in raw_deps}
        for raw_extras in _nested_item(
            self.cfg,
            "project.optional-dependencies",
            dict,
        ).values():
            deps |= {canonical(raw_extra) for raw_extra in raw_extras}
        return frozenset(deps)


def pkg(module: str) -> str:
    """Get the installable module name from an import or package name statement."""
    return module.split(".")[0].replace("-", "_")


T = TypeVar("T")


def _nested_item(obj: dict[str, Any], keys: str, class_: type[T]) -> T:
    """Get items from a nested dictionary where the keys are dot-separated."""
    for a in keys.split("."):
        if a not in obj:
            return class_()
        obj = obj[a]
    if not isinstance(obj, class_):
        msg = f"Expected {class_} but got {type(obj)}"
        raise TypeError(msg)
    return cast("T", obj)


def _get_pyproject_path(path: Path) -> Path:
    for p in chain([path], path.resolve().parents):
        if (p / _PYPROJECT_TOML).exists():
            return p / _PYPROJECT_TOML
    msg = f"Could not find {_PYPROJECT_TOML} file within path hierarchy"
    raise FileNotFoundError(msg)
